
# PLEASE READ DISCLAIMER
#
#    This class demonstrates sample IxNetwork REST API usage for
#    demo and reference purpose only.
#    It is subject to change for updates without warning.
#
# REQUIREMENTS
#    - Python 2.7 (Supports Python 2 and 3)
#    - Python modules: requests
#

from __future__ import absolute_import, print_function, division
import os, re, sys, requests, json, time, subprocess, traceback, time, datetime, platform

class IxNetRestApiException(Exception):
    def __init__(self, msg=None):
        if platform.python_version().startswith('3'):
            super().__init__(msg)

        if platform.python_version().startswith('2'):
            super(IxNetRestApiException, self). __init__(msg)

        if Connect.robotStdout is not None:
            Connect.robotStdout.log_to_console(msg)

        showErrorMsg = '\nIxNetRestApiException error: {0}\n\n'.format(msg)
        print(showErrorMsg)
        if Connect.enableDebugLogFile:
            with open(Connect.debugLogFile, 'a') as restLogFile:
                restLogFile.write(showErrorMsg)

class Connect:
    # For IxNetRestApiException
    debugLogFile = None
    enableDebugLogFile = False
    robotStdout = None

    def __init__(self, apiServerIp=None, serverIpPort=None, serverOs='windows', connectToLinuxChassisIp=None, manageSessionMode=False,
                 webQuickTest=False, username=None, password='admin', licenseServerIp=None, licenseMode=None, licenseTier=None,
                 deleteSessionAfterTest=True, verifySslCert=False, includeDebugTraceback=True, sessionId=None, httpsSecured=False,
                 apiKey=None, generateLogFile=True, robotFrameworkStdout=False,proxies=None, trustEnv=True):
        """
        Description
           Initializing default parameters and making a connection to the API server

        Notes
            Starting IxNetwork 8.50, https will be enforced even for Windows connection.
            If you still want to use http, you need to add -restInsecure to the IxNetwork.exe appliaction under "target".

        Examples
            Right click on "IxNetwork API server", select properties and under target
            ixnetwork.exe -restInsecure -restPort 11009 -restOnAllInterfaces -tclPort 8009

        Parameters
           apiServerIp: (str): The API server IP address.
           serverIpPort: (str): The API server IP address socket port.
           serverOs: (str): windows|windowsConnectionMgr|linux
           connectToLinuxChassis: (str): Connect to a Linux OS chassis IP address.
           webQuickTest: (bool): True: Using IxNetwork Web Quick Test. Otherwise, using IxNetwork.
           includeDebugTraceback: (bool):
                                   True: Traceback messsages are included in raised exceptions.
                                   False: No traceback.  Less verbose for debugging.
           username: (str): The login username. For Linux API server only.
           password: (str): The login password. For Linux API server only.
           licenseServerIp: (str): The license server IP address.
           licenseMode: (str): subscription | perpetual | mixed
           licenseTier: (str): tier1 | tier2 | tier3
           deleteSessionAfterTest: (bool): True: Delete the session.
                                           False: Don't delete the session.
           verifySslCert: (str): Optional: Include your SSL certificate for added security.
           httpsSecured: (bool): This parameter is only used by Connection Mgr when user wants to connect to
                                 an existing session.
                                 True = IxNetwork ReST API server is using HTTPS.
                                 This parameter must also include sessionId and serverIpPort=<the ssl port number>

           serverOs: (str): Defaults to windows. windows|windowsConnectionMgr|linux.
           includeDebugTraceback: (bool): True: Include tracebacks in raised exceptions.
           sessionId: (str): The session ID on the Linux API server or Windows Connection Mgr to connect to.
           apiKey: (str): The Linux API server user account API-Key to use for the sessionId connection.
           generateLogFile: True|False|<log file name>.  If you want to generate a log file, provide
                                the log file name.
                                True = Then the log file default name is ixNetRestApi_debugLog.txt
                                False = Disable generating a log file.
                                <log file name> = The full path + file name of the log file to create.
           robotFrameworkStdout: (bool):  True = Print to stdout.
           httpInsecure: (bool): This parameter is only for Windows connections.
                                     True: Using http.  False: Using https.
                                     Starting 8.50: IxNetwork defaults to use https.
                                     If you are using versions prior to 8.50, it needs to be a http connection.
                                     In this case, set httpInsecure=True.

        Notes
            Class attributes
               self._session: The requests initial Session().
               self.serverOs: windows|windowsConnectionMgr|linux

               self.httpHeader: http://{apiServerIp}:{port}
               self.sessionId : http://{apiServerIp}:{port}/api/v1/sessions/{id}
               self.sessionUrl: http://{apiServerIp}:{port}/api/v1/sessions/{id}/ixnetwork
               self.headlessSessionId: /api/v1/sessions/{id}
               self.apiSessionId:      /api/v1/sessions/{id}/ixnetwork

               self.jsonHeader: The default header: {"content-type": "application/json"}
               self.apiKey: For Linux API server only. Automatically provided by the server when login
                            successfully authenticated.
                            You could also provide an API-Key to connect to an existing session.
                            Get the API-Key from the Linux API server user account.

        Examples:
           Steps to connect to Linux API server steps:
               1> POST: https://{apiServerIp}/api/v1/auth/session
                  DATA: {"username": "admin", "password": "admin"}
                  HEADERS: {'content-type': 'application/json'}

               2> POST: https://{apiServerIp:{port}/api/v1/sessions
                  DATA: {"applicationType": "ixnrest"}
                  HEADERS: {'content-type': 'application/json', 'x-api-key': 'd9f4da46f3c142f48dddfa464788hgee'}

               3> POST: https://{apiServerIp}:443/api/v1/sessions/4/operations/start
                  DATA: {}
                  HEADERS: {'content-type': 'application/json', 'x-api-key': 'd9f4da46f3c142f48dddfa464788hgee'}

               sessionId = https://{apiServerIp}:443/api/v1/sessions/{id}

           Steps to connect to Linux Web Quick Test:
               1> POST: https://{apiServerIp}:443/api/v1/auth/session
                  DATA: {"username": "admin", "password": "admin"}
                  HEADERS: {'content-type': 'application/json'}

               2> POST: https://{apiServeIp}:443/api/v1/sessions
                  DATA: {'applicationType': 'ixnetwork'}

               3> POST: https://{apiServerIp}:443/api/v1/sessions/2/operations/start
                  DATA: {'applicationType': 'ixnetwork'}

            sessionId = https://{apiServerIp}/ixnetworkweb/api/v1/sessions/{id}

           Notes
              To connect to an existing configuration.
                 Windows: Nothing special to include. The session ID is always "1".
                 Linux API server: Include the api-key and sessionId that you want to connect to.
                 Windows Connection Manager: Include just the sessionId: For example: 8021.
        """
        from requests.exceptions import ConnectionError
        from requests.packages.urllib3.connection import HTTPConnection

        # Disable SSL warnings
        requests.packages.urllib3.disable_warnings()

        # Disable non http connections.
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        self._session = requests.Session()

        self.httpScheme = 'http' ;# This will change to https in createWindowsSession and connectToLinuxApiServer
        if httpsSecured:
            # For Windows Connection Mgr only.
            # When creating a new session, there is no way to know by doing a POST for a new session to
            # understand if it's for http or https.  You must enter https for the POST to create a new session.
            self.httpScheme = 'https'

        self.serverOs = serverOs ;# windows|windowsConnectionMgr|linux
        self.jsonHeader = {"content-type": "application/json"}
        self.username = username
        self.password = password
        self.apiKey = apiKey
        self.verifySslCert = verifySslCert
        self.linuxApiServerIp = apiServerIp
        self.manageSessionMode = manageSessionMode
        self.apiServerPort = serverIpPort
        self.webQuickTest = webQuickTest
        self.generateLogFile = generateLogFile
        self.robotFrameworkStdout = robotFrameworkStdout
        self.connectToLinuxChassisIp = connectToLinuxChassisIp
        self.proxies = proxies
        self.trustEnv = trustEnv

        if generateLogFile:
            if generateLogFile == True:
                # Default the log file name
                self.restLogFile = 'ixNetRestApi_debugLog.txt'
                Connect.enableDebugLogFile = True
                Connect.debugLogFile = self.restLogFile

            if type(generateLogFile) != bool:
                self.restLogFile = generateLogFile

            # Instantiate a new log file here.
            with open(self.restLogFile, 'w') as restLogFile:
                restLogFile.write('Date: {0}\nTime: {1}\n\n'.format(self.getDate, self.getTime))

        # Make Robot print to stdout
        if self.robotFrameworkStdout:
            from robot.libraries.BuiltIn import _Misc
            self.robotStdout = _Misc()
            Connect.robotStdout = self.robotStdout

        if connectToLinuxChassisIp:
            self.connectToLinuxIxosChassis(self.connectToLinuxChassisIp, self.username, self.password)
            return

        if serverOs == 'windows':
            self.createWindowsSession(apiServerIp, serverIpPort)

        if serverOs == 'windowsConnectionMgr':
            # User connecting to existing sessionId
            # 
            # Starting 8.50, IxNetwork API server supports https on Windows.
            # If connecting to an existing session with the api server set to https::
            #     - Pass in apiServerIp, serverIpPort, httpsSecured and  sessionId
            #     - Although the serverIpPort default is 443 for https, this could change in the future.
            
            if sessionId:
                url = '{0}://{1}:{2}/api/v1/sessions/{3}'.format(self.httpScheme, apiServerIp, serverIpPort, str(sessionId))
                try:
                    response = self._session.request('GET', url, verify=self.verifySslCert, allow_redirects=False)
                    if '3' in str(response.status_code):
                        self.httpScheme = 'https'
                        # Here, needs to set to use https.
                        url = '{0}://{1}:{2}/api/v1/sessions/{3}'.format(self.httpScheme, apiServerIp, serverIpPort, str(sessionId))

                except requests.exceptions.RequestException as errMsg:
                    errMsg = 'Connecting to existing config failed on a GET: {0}'.format(errMsg)
                    raise IxNetRestApiException(errMsg)

                self.logInfo('Connecting to existing session: {}'.format(url))
                self.sessionUrl = url + '/ixnetwork'
                self.sessionId = '{0}://{1}:{2}/api/v1/sessions/{3}'.format(self.httpScheme, apiServerIp,
                                                                            serverIpPort, str(sessionId))
                self.apiSessionId = '/api/v1/sessions/{0}/ixnetwork'.format(str(sessionId))
                self.httpHeader = self.sessionUrl.split('/api')[0]
            else:
                # Create a new session
                self.createWindowsSession(apiServerIp, serverIpPort)

        if serverOs == 'linux':
            if self.apiServerPort == None:
                self.apiServerPort = 443

            # Connect to no session ID. Connects to a Linux API server to get and delete open sessions.
            if self.manageSessionMode:
                if self.apiKey == None:
                    response = self.post('https://{}:{}/api/v1/auth/session'.format(self.linuxApiServerIp, self.apiServerPort),
                                         data={"username": "admin", "password": "admin"},
                                         headers={'content-type': 'application/json'})

                self.jsonHeader = {'content-type': 'application/json', 'x-api-key': self.apiKey}
                self.sessionUrl = 'https://{0}:{1}/ixnetworkweb/api/v1/sessions'.format(self.linuxApiServerIp, self.apiServerPort)
                return

            # Connect to an existing session on the Linux API server
            if apiKey and sessionId:
                self.sessionId = 'https://{0}:{1}/api/v1/sessions/{2}'.format(self.linuxApiServerIp,
                                                                              self.apiServerPort, str(sessionId))
                self.sessionUrl = self.sessionId + '/ixnetwork'
                self.httpHeader = self.sessionUrl.split('/ixnetworkweb')[0]
                self.jsonHeader = {'content-type': 'application/json', 'x-api-key': self.apiKey}

                response = self.get('https://{0}:{1}/api/v1/sessions/{2}'.format(self.linuxApiServerIp, self.apiServerPort, str(sessionId)))

                if self.webQuickTest == False:
                    # https://192.168.70.108/ixnetworkweb/api/v1/sessions/4
                    self.sessionId = response.json()['links'][0]['href']
                    # Remove the redirect /ixnetworkweb from the URL. IxNetwork 8.50 will resolve this.
                    self.sessionId = self.sessionId.replace('ixnetworkweb/', '')

                    # https://192.168.70.108/ixnetworkweb/api/v1/sessions/4/ixnetork
                    self.sessionUrl = self.sessionId + '/ixnetwork'

                    # /api/v1/sessions/4/ixnetwork
                    match = re.match('.*(/api.*)', self.sessionId)
                    self.apiSessionId = match.group(1) + '/ixnetwork'

                    # https://10.10.10.1:443
                    matchHeader = re.match('(https://[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+(:[0-9]+)?)', self.sessionId)
                    self.httpHeader = matchHeader.group(1)

                if self.webQuickTest:
                    self.sessionId = 'https://{0}:{1}/ixnetwork/api/v1/sessions/{2}'.format(self.linuxApiServerIp,
                                                                                               self.apiServerPort, str(sessionId))
                    self.sessionUrl = self.sessionId
                    self.httpHeader = self.sessionUrl.split('/ixnetworkweb')[0]

            # Create new session: connectToLinuxApiServer API knows whether to create a new session or connect to an existing
            #                     session by looking at the self.apiKey.
            if apiKey == None and sessionId == None:
                self.connectToLinuxApiServer(self.linuxApiServerIp, self.apiServerPort, username=username, password=password, verifySslCert=verifySslCert)

            if licenseServerIp or licenseMode or licenseTier:
                self.configLicenseServerDetails(licenseServerIp, licenseMode, licenseTier)

        # For Linux API Server and Windoww Connection Mgr only: Delete the session when script is done if deleteSessionAfterTest = True.
        self.deleteSessionAfterTest = deleteSessionAfterTest

        match = re.match('http.*(/api.*/sessions/[0-9]+)/ixnetwork', self.sessionUrl)
        self.headlessSessionId = match.group(1)

        if includeDebugTraceback == False:
            sys.tracebacklimit = 0

    def get(self, restApi, data={}, stream=False, silentMode=False, ignoreError=False):
        """
        Description
            A HTTP GET function to send REST APIs.

        Parameters
           restApi: (str): The REST API URL.
           data: (dict): The data payload for the URL.
           silentMode: (bool):  To display on stdout: URL, data and header info.
           ignoreError: (bool): True: Don't raise an exception.  False: The response will be returned.

        Syntax
            /api/v1/sessions/1/ixnetwork/operations
        """
        if silentMode is False:
            self.logInfo('\n\tGET: {0}\n\tHEADERS: {1}'.format(restApi, self.jsonHeader))

        ses = requests.Session()
        ses.headers=self.jsonHeader
        ses.verify=self.verifySslCert
        ses.allow_redirects=True
        ses.trust_env=self.trustEnv

        if stream:
            ses.stream=True

        if self.proxies != None:
            ses.proxies=self.proxies

        try:
            response = ses.get(restApi)

            if silentMode is False:
                for redirectStatus in response.history:
                    if '307' in str(response.history):
                        self.logInfo('\t{0}: {1}'.format(redirectStatus, response.url), timestamp=False)

                self.logInfo('\tSTATUS CODE: {0}'.format(response.status_code), timestamp=False)

            if not str(response.status_code).startswith('2'):
                if ignoreError == False:
                    if 'message' in response.json() and response.json()['messsage'] != None:
                        self.logWarning('\n%s' % response.json()['message'])

                    errMsg = 'GET Exception error: {0}'.format(response.text)
                    raise IxNetRestApiException(errMsg)

            return response

        except requests.exceptions.RequestException as errMsg:
            errMsg = 'GET Exception error: {0}'.format(errMsg)
            raise IxNetRestApiException(errMsg)

    def post(self, restApi, data={}, headers=None, silentMode=False, noDataJsonDumps=False, ignoreError=False):
        """
        Description
           A HTTP POST function to create and start operations.

        Parameters
           restApi: (str): The REST API URL.
           data: (dict): The data payload for the URL.
           headers: (str): The special header to use for the URL.
           silentMode: (bool):  To display on stdout: URL, data and header info.
           noDataJsonDumps: (bool): True: Use json dumps. False: Accept the data as-is.
           ignoreError: (bool): True: Don't raise an exception.  False: The response will be returned.
        """
        ses = requests.Session()
        ses.verify=self.verifySslCert
        ses.allow_redirects=True
        ses.trust_env=self.trustEnv
  
        if headers != None:
            originalJsonHeader = self.jsonHeader
            self.jsonHeader = headers
            ses.headers=headers
        else:
            ses.headers=self.jsonHeader

        if self.proxies != None:
            ses.proxies=self.proxies

        if noDataJsonDumps == True:
            data = data
        else:
            data = json.dumps(data)

        if silentMode == False:
            self.logInfo('\n\tPOST: {0}\n\tDATA: {1}\n\tHEADERS: {2}'.format(restApi, data, ses.headers))

        try:
            response = ses.post(restApi, data=data)

            # 200 or 201
            if silentMode == False:
                for redirectStatus in response.history:
                    if '307' in str(response.history):
                        self.logInfo('\t{0}: {1}'.format(redirectStatus, response.url), timestamp=False)

                self.logInfo('\tSTATUS CODE: %s' % response.status_code, timestamp=False)

            if str(response.status_code).startswith('2') == False:
                if ignoreError == False:
                    if 'errors' in response.json():
                        errMsg = 'POST Exception error: {0}\n'.format(response.json()['errors'])
                        raise IxNetRestApiException(errMsg)

                    raise IxNetRestApiException('POST error: {0}\n'.format(response.text))

            return response

        except requests.exceptions.RequestException as errMsg:
            errMsg = 'POST Exception error: {0}'.format(errMsg)
            raise IxNetRestApiException(errMsg)

    def patch(self, restApi, data={}, silentMode=False, ignoreError=False):
        """
        Description
           A HTTP PATCH function to modify configurations.

        Parameters
           restApi: (str): The REST API URL.
           data: (dict): The data payload for the URL.
           silentMode: (bool):  To display on stdout: URL, data and header info.
        """
        ses = requests.Session()
        ses.verify=self.verifySslCert
        data = json.dumps(data)
        ses.headers = self.jsonHeader
        ses.trust_env=self.trustEnv

        if self.proxies != None:
            ses.proxies=self.proxies

        if silentMode == False:
            self.logInfo('\n\tPATCH: {0}\n\tDATA: {1}\n\tHEADERS: {2}'.format(restApi, data, self.jsonHeader))

        try:
            response = ses.patch(restApi, data=data)

            if silentMode == False:
                for redirectStatus in response.history:
                    if '307' in str(response.history):
                        self.logInfo('\t{0}: {1}'.format(redirectStatus, response.url), timestamp=False)

                self.logInfo('\tSTATUS CODE: %s' % response.status_code, timestamp=False)

                if ignoreError == False:
                    if not str(response.status_code).startswith('2'):
                        if response.json() and  'errors' in response.json():
                            errMsg = 'PATCH Exception error: {0}\n'.format(response.json()['errors'])
                            raise IxNetRestApiException('PATCH error: {0}\n'.format(errMsg))

            return response

        except requests.exceptions.RequestException as errMsg:
            errMsg = 'PATCH Exception error: {0}\n'.format(errMsg)
            raise IxNetRestApiException(errMsg)

    def options(self, restApi, data={}, silentMode=False, ignoreError=False):
        """
        Description
            A HTTP OPTIONS function to send REST APIs.

        Parameters
           restApi: (str): The REST API URL.
           silentMode: (bool):  To display on stdout: URL, data and header info.
           ignoreError: (bool): True: Don't raise an exception.  False: The response will be returned.

        """
        ses = requests.Session()
        ses.verify=self.verifySslCert
        ses.headers = self.jsonHeader
        ses.trust_env=self.trustEnv

        if self.proxies != None:
            ses.proxies=self.proxies

        if silentMode is False:
            self.logInfo('\n\tOPTIONS: {0}\n\tHEADERS: {1}'.format(restApi, self.jsonHeader))

        try:
            # For binary file
            response = ses.get(restApi, data=data)

            if silentMode is False:
                for redirectStatus in response.history:
                    if '307' in str(response.history):
                        self.logInfo('\t{0}: {1}'.format(redirectStatus, response.url), timestamp=False)

                self.logInfo('\tSTATUS CODE: {0}'.format(response.status_code), timestamp=False)

            if not str(response.status_code).startswith('2'):
                if ignoreError == False:
                    if 'message' in response.json() and response.json()['messsage'] != None:
                        self.logWarning('\n%s' % response.json()['message'])

                    errMsg = 'OPTIONS Exception error: {0}'.format(response.text)
                    raise IxNetRestApiException(errMsg)
            return response

        except requests.exceptions.RequestException as errMsg:
            errMsg = 'OPTIONS Exception error: {0}'.format(errMsg)
            raise IxNetRestApiException(errMsg)

    def delete(self, restApi, data={}, headers=None):
        """
        Description
           A HTTP DELETE function to delete the session.
           For Linux and Windows Connection Mgr API server only.

        Paramters
           restApi: (str): The REST API URL.
           data: (dict): The data payload for the URL.
           headers: (str): The headers to use for the URL.
        """
        ses = requests.Session()
        ses.verify=self.verifySslCert
        ses.verify=self.verifySslCert
        ses.data = json.dumps(data)

        if headers != None:
            ses.headers = headers
        else:
            ses.headers = self.jsonHeader

        if self.proxies != None:
            ses.proxies=self.proxies

        self.logInfo('\n\tDELETE: {0}\n\tDATA: {1}\n\tHEADERS: {2}'.format(restApi, data, ses.headers))

        try:
            response = ses.delete(restApi, data=data)

            for redirectStatus in response.history:
                if '307' in str(response.history):
                    self.logInfo('\t{0}: {1}'.format(redirectStatus, response.url), timestamp=False)

            self.logInfo('\tSTATUS CODE: %s' % response.status_code, timestamp=False)
            if not str(response.status_code).startswith('2'):
                self.showErrorMessage()
                errMsg = 'DELETE Exception error: {0}\n'.format(response.text)
                self.logError(errMsg)
                raise IxNetRestApiException(errMsg)
            return response

        except requests.exceptions.RequestException as errMsg:
            errMsg = 'DELETE Exception error: {0}\n'.format(errMsg)
            self.logError(errMsg)
            raise IxNetRestApiException(errMsg)

    def getDate(self):
        dateAndTime = str(datetime.datetime.now()).split(' ')
        return dateAndTime[0]

    def getTime(self):
        dateAndTime = str(datetime.datetime.now()).split(' ')
        return dateAndTime[1]

    def getSelfObject(self):
        """
        Description
           For Robot Framework support only.

        Return
           The instance object.
        """
        return self

    def createWindowsSession(self, ixNetRestServerIp, ixNetRestServerPort='11009'):
        """
        Description
           Connect to a Windows IxNetwork API Server. This is
           for both Windows and Windows server with IxNetwork Connection Manager.
           This will set up the session URL to use throughout the test.

        Parameter
          ixNetRestServerIp: (str): The Windows IxNetwork API Server IP address.
          ixNetRestServerPort: (str): Default: 11009.  Provide a port number to connect to.
                               On a Linux API Server, a socket port is not needed. State "None".
        """
        # Handle __import__(IxNetRestApi) to not error out
        if ixNetRestServerIp == None: return

        url = 'http://{0}:{1}/api/v1/sessions'.format(ixNetRestServerIp, ixNetRestServerPort)

        if self.serverOs == 'windowsConnectionMgr':
            try:
                self.logInfo('Please wait while IxNetwork Connection Mgr starts up an IxNetwork session...')
                self.logInfo('\t{}'.format(url), timestamp=False)
                response = self._session.request('POST', url, data={}, verify=self.verifySslCert, allow_redirects=True)

            except requests.exceptions.RequestException as errMsg:
                errMsg = 'Creating new session failed: {0}'.format(errMsg)
                raise IxNetRestApiException(errMsg)

            sessionIdNumber = response.json()['links'][0]['href'].split('/')[-1]
            response = self._session.request('GET', url+'/'+str(sessionIdNumber), verify=self.verifySslCert, allow_redirects=False)

            if str(response.status_code).startswith('3'):
                # >= 8.50
                self.httpScheme = 'https'
                # https://192.168.70.3:443/api/v1/sessions/8020
                self.sessionId = response.headers['location']
                match = re.match('https://.*:([0-9]+)/api.*', self.sessionId)
                ixNetRestServerPort = match.group(1)
            else:
                # < 8.50
                self.sessionId = url + '/'+str(sessionIdNumber)

            self.sessionUrl = '{httpScheme}://{apiServer}:{port}/api/v1/sessions/{id}/ixnetwork'.format(
                httpScheme=self.httpScheme,
                apiServer=ixNetRestServerIp,
                port=ixNetRestServerPort,
                id=sessionIdNumber)

            counterStop = 10
            for counter in range(1, counterStop+1):
                response = self.get(self.sessionId)
                if type(response.json()) == list:
                    currentState = response.json()[0]['state']
                else:
                    currentState = response.json()['state']

                self.logInfo('\n\tNew Windows session current state: {0}'.format(currentState), timestamp=False)
                if currentState != 'ACTIVE' and counter < counterStop:
                    self.logInfo('\tWaiting {0}/{1} seconds'.format(counter, counterStop), timestamp=False)
                    time.sleep(1)

                if currentState != 'ACTIVE' and counter == counterStop:
                    raise IxNetRestApiException('New Windows session state failed to become ACTVIE state')

                if currentState == 'ACTIVE' and counter < counterStop:
                    break

            # Version < 8.50 requires more waiting time. And we cannot verify the version number at this point.
            time.sleep(25)

        if self.serverOs == 'windows':
            # windows sessionId is always 1 because it only supports one session.
            self.sessionUrl = url + '/1/ixnetwork'

            try:
                self.logInfo('\nVerifying API server connection: {}'.format(self.sessionUrl))
                response = self._session.request('GET', self.sessionUrl, allow_redirects=False, verify=self.verifySslCert)
                if '307' in str(response.status_code) and response.headers['location'].startswith('https'):
                    # Overwrite the sessionUrl with the redirected https URL
                    self.sessionUrl = response.headers['location']

            except requests.exceptions.RequestException as errMsg:
                errMsg = 'Connection failed: {0}'.format(errMsg)
                raise IxNetRestApiException(errMsg)

            # http://192.168.70.127:11009/api/v1/sessions/1
            self.sessionId = self.sessionUrl.split('/ixnetwork')[0]

        # http://192.168.70.127:11009
        self.httpHeader = self.sessionUrl.split('/api')[0]

        # /api/v1/sessions/{id}/ixnetwork
        self.apiSessionId = '/api/v1/sessions/{0}/ixnetwork'.format(str(self.sessionId))

        self.get(self.sessionUrl)

    def deleteSession(self):
        """
        Description
           Delete the instance session ID. For Linux and Windows Connection Manager only.
        """
        if self.deleteSessionAfterTest:
            self.delete(self.sessionId)

    def logInfo(self, msg, end='\n', timestamp=True):
        """
        Description
           An internal function to print info to stdout

        Parameters
           msg: (str): The message to print.
        """
        currentTime = self.getTime()

        if timestamp:
            msg = '\n' + currentTime + ': ' + msg
        else:
            # No timestamp and no newline are mainly for verifying states and status
            msg = msg

        print('{0}'.format(msg), end=end)
        if self.generateLogFile:
            with open(self.restLogFile, 'a') as restLogFile:
                restLogFile.write(msg+end)

        if self.robotFrameworkStdout:
            self.robotStdout.log_to_console(msg)

    def logWarning(self, msg, end='\n', timestamp=True):
        """
        Description
           An internal function to print warnings to stdout.

        Parameter
           msg: (str): The message to print.
        """
        currentTime = self.getTime()

        if timestamp:
            msg = '\n{0}: Warning: {1}'.format(currentTime, msg)
        else:
            # No timestamp and no newline are mainly for verifying states and status
            msg = msg

        print('{0}'.format(msg), end=end)
        if self.generateLogFile:
            with open(self.restLogFile, 'a') as restLogFile:
                restLogFile.write('Warning: '+msg+end)

        if self.robotFrameworkStdout:
            self.robotStdout.log_to_console(msg)

    def logError(self, msg, end='\n', timestamp=True):
        """
        Description
           An internal function to print error to stdout.

        Parameter
           msg: (str): The message to print.
        """
        currentTime = self.getTime()

        if timestamp:
            msg = '\n{0}: Error: {1}'.format(currentTime, msg)
        else:
            # No timestamp and no newline are mainly for verifying states and status
            msg = '\nError: {0}'.format(msg)

        print('{0}'.format(msg), end=end)
        if self.generateLogFile:
            with open(self.restLogFile, 'a') as restLogFile:
                restLogFile.write('Error: '+msg+end)

        if self.robotFrameworkStdout:
            self.robotStdout.log_to_console(msg)

    def getIxNetworkVersion(self):
        """
        Description
           Get the IxNetwork version.

        Syntax
            GET: /api/v1/sessions/{id}/globals
        """
        response = self.get(self.sessionUrl+'/globals', silentMode=True)
        return response.json()['buildNumber']

    def getAllSessionId(self):
        """
        Show all opened session IDs.

        Return
           A list of opened session IDs.

           {4: {'startedOn': '2018-10-06 12:09:18.333-07:00',
              'state': 'Active',
              'subState': 'Ready',
              'userName': 'admin'},
            5: {'startedOn': '2018-10-06 18:49:05.691-07:00',
              'state': 'Active',
              'subState': 'Ready',
              'userName': 'admin'}
           }
        """
        response = self.get(self.httpHeader+'/api/v1/sessions', silentMode=True)
        sessionId = {}
        for eachSessionId in response.json():
            sessionId.update({eachSessionId['id']: {'userName': eachSessionId['userName'],
                                                    'createdOn': eachSessionId['createdOn'].replace('T', ' '),
                                                    'state': eachSessionId['state'],
                                                    'subState': eachSessionId['subState']
                                                }
                          })

            print('\nSessionId: {}'.format(eachSessionId['id']))
            print('   UserName: {}'.format(eachSessionId['userName']))
            print('   StartedOn: {}'.format(eachSessionId['createdOn'].replace('T', ' ')))
            print('   State: {}'.format(eachSessionId['state']))
            print('   SubState: {}'.format(eachSessionId['subState']))

        print()
        return sessionId

    def showErrorMessage(self, silentMode=False):
        """
        Description
           Show all the error messages from IxNetwork.

        Parameter
          silentMode: (bool): True: Don't print the REST API on stdout.

        Syntax
            GET: /api/v1/sessions/{id}/globals/appErrors/error
        """
        errorList = []
        response = self.get(self.sessionUrl+'/globals/appErrors/error', silentMode=silentMode)
        print()
        for errorId in response.json():
            if errorId['errorLevel'] == 'kError':
                print('CurrentErrorMessage: {0}'.format(errorId['name']))
                print('\tDescription: {0}'.format(errorId['lastModified']))
                errorList.append(errorId['name'])
        print()
        return errorList

    def waitForComplete(self, response='', url='', silentMode=False, ignoreException=False, timeout=120):
        """
        Description
           Wait for an operation progress to complete.

        Parameters
           response: (json response/dict): The POST action response.  Generally, after an /operations action.
                         Such as /operations/startallprotocols, /operations/assignports.
           silentMode: (bool):  If True, display info messages on stdout.
           ignoreException: (bool): ignoreException is for assignPorts.  Don't want to exit test.
                            Verify port connectionStatus for: License Failed and Version Mismatch to report problem immediately.

           timeout: (int): The time allowed to wait for success completion in seconds.
        """
        if silentMode == False:
            self.logInfo('\nwaitForComplete:', timestamp=False)
            self.logInfo("\tState: %s " % response.json()["state"], timestamp=False)

        if response.json() == []:
            raise IxNetRestApiException('waitForComplete: response is empty.')

        if response.json() == '' or response.json()['state'] == 'SUCCESS':
            return response

        if 'errors' in response.json():
            raise IxNetRestApiException(response.json()["errors"][0])

        if response.json()['state'] in ["ERROR", "EXCEPTION"]:
            raise IxNetRestApiException('WaitForComplete: STATE=%s: %s' % (response.json()['state'], response.text))

        for counter in range(1,timeout+1):
            response = self.get(url, silentMode=True)
            state = response.json()["state"]
            if silentMode == False:
                if state != 'SUCCESS':
                    self.logInfo("\tState: {0}: Wait {1}/{2} seconds".format(state, counter, timeout), timestamp=False)
                if state == 'SUCCESS':
                    self.logInfo("\tState: {0}".format(state), timestamp=False)

            if counter < timeout and state in ["IN_PROGRESS", "down"]:
                time.sleep(1)
                continue

            if counter < timeout and state in ["ERROR", "EXCEPTION"]:
                # ignoreException is for assignPorts.  Don't want to exit test.
                # Verify port connectionStatus for: License Failed and Version Mismatch to report problem immediately.
                if ignoreException:
                    return response
                raise IxNetRestApiException(response.text)

            if counter < timeout and state == 'SUCCESS':
                return response

            if counter == timeout and state != 'SUCCESS':
                if ignoreException:
                    return response
                raise IxNetRestApiException('waitForComplete failed: %s' % response.json())

    def connectToLinuxIxosChassis(self, chassisIp, username, password):
        url = 'https://{0}/platform/api/v1/auth/session'.format(chassisIp)
        response = self.post(url, data={'username': username, 'password': password})
        self.apiKey = response.json()['apiKey']
        self.jsonHeader = {'content-type': 'application/json', 'x-api-key': self.apiKey}

        # userAccountUrl: https://{ip}/platform/api/v1/auth/users/{id}
        self.userSessionId = response.json()['userAccountUrl']

        self.ixosHeader = 'https://{0}/chassis/api/v2/ixos'.format(chassisIp)
        self.diagnosticsHeader = 'https://{0}/chassis/api/v1/diagnostics'.format(chassisIp)
        self.authenticationHeader = 'https://{0}/chassis/api/v1/auth'.format(chassisIp)
        self.sessionUrl = self.ixosHeader

    def connectToLinuxApiServer(self, linuxServerIp, linuxServerIpPort, username='admin', password='admin', verifySslCert=False):
        """
        Description
           Connect to a Linux API server.

        Parameters
           linuxServerIp: (str): The Linux API server IP address.
           username: (str): Login username. Default = admin.
           password: (str): Login password. Default = admin.
           verifySslCert: (str): Defalt: None.  The SSL Certificate for secure access verification.

       Syntax
            POST: /api/v1/auth/session
        """
        self.verifySslCert = verifySslCert

        if self.apiKey is None:
            # 1: Connect to the Linux API server
            url = 'https://{0}:{1}/api/v1/auth/session'.format(linuxServerIp, linuxServerIpPort)
            self.logInfo('connectToLinuxApiServer: %s' % url)
            response = self.post(url, data={'username': username, 'password': password}, ignoreError=True)
            if not str(response.status_code).startswith('2'):
                raise IxNetRestApiException('\nLogin username/password failed\n')
            self.apiKey = response.json()['apiKey']

            url = 'https://{0}:{1}/api/v1/sessions'.format(linuxServerIp, linuxServerIpPort)
            if self.webQuickTest == False:
                data = {'applicationType': 'ixnrest'}
            if self.webQuickTest == True:
                data = {'applicationType': 'ixnetwork'}

            self.jsonHeader = {'content-type': 'application/json', 'x-api-key': self.apiKey}
            self.logInfo('linuxServerCreateSession')
            response = self.post(url, data=data, headers=self.jsonHeader)

            self.sessionIdNumber = response.json()['id']
            response = self.get(url+'/'+str(self.sessionIdNumber))

            # https://192.168.70.108/ixnetworkweb/api/v1/sessions/7
            self.sessionId = response.json()['links'][0]['href']
            # Remove the redirect /ixnetworkweb from the URL. IxNetwork 8.50 will resolve this.
            self.sessionId = self.sessionId.replace('ixnetworkweb/', '')

            # https://10.10.10.1:443
            matchHeader = re.match('(https://[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+(:[0-9]+)?)', self.sessionId)
            self.httpHeader = matchHeader.group(1)

            if ':' not in self.httpHeader:
                self.httpHeader + '/' + linuxServerIpPort

            self.sessionUrl = self.sessionId+'/ixnetwork'

            # /api/v1/sessions/4/ixnetwork
            match = re.match('.*(/api.*)', self.sessionId)
            self.apiSessionId = match.group(1) + '/ixnetwork'

            # 3: Start the new session
            response = self.post(self.sessionId+'/operations/start')
            if self.linuxServerWaitForSuccess(response.json()['url'], timeout=60) == 1:
                raise IxNetRestApiException

            if self.webQuickTest == True:
                self.sessionId = 'https://{0}/ixnetworkweb/api/v1/sessions/{1}'.format(linuxServerIp, self.sessionIdNumber)
                self.sessionUrl = 'https://{0}/ixnetworkweb/api/v1/sessions/{1}/ixnetwork'.format(linuxServerIp, self.sessionIdNumber)
                self.httpHeader = self.sessionUrl.split('/api')[0]

        # If an API-Key is provided, then verify the session ID connection.
        if self.apiKey:
            self.get(self.sessionId)

    def linuxServerGetGlobalLicense(self, linuxServerIp):
        """
        Description
           Get the global license server details from the Linux API server.

        Paramters
           linuxServerIp: (str): The IP address of the Linux API server.

        Syntax
            GET: /api/v1/sessions/9999/ixnetworkglobals/license
        """
        staticUrl = 'https://{linuxServerIp}/api/v1/sessions/9999/ixnetworkglobals/license'.format(linuxServerIp=linuxServerIp)
        self.logInfo('linuxServerGetGlobalLicense: %s ' % linuxServerIp)
        response = self.get(staticUrl, silentMode=False)
        licenseServerIp = response.json()['servers'][0]
        licenseServerMode = response.json()['mode']
        licenseServerTier = response.json()['tier']
        self.logInfo('linuxServerGetGlobalLicenses:')
        self.logInfo('\t%s' % licenseServerIp)
        self.logInfo('\t%s' % licenseServerMode)
        self.logInfo('\t%s' % licenseServerTier)
        return licenseServerIp,licenseServerMode,licenseServerTier

    def configLicenseServerDetails(self, licenseServer=None, licenseMode=None, licenseTier=None):
        """
        Description
           Configure license server details: license server IP, license mode and license tier.

        Parameters
           licenseServer: (str): License server IP address(s) in a list.
           licenseMode: (str): subscription | perpetual | mixed
           licenseTier: (str): tier1 | tier2 | tier3 ...

        Syntax
           PATCH: /api/v1/sessions/{id}/ixnetwork/globals/licensing
        """
        # Each new session requires configuring the new session's license details.
        data = {}
        if licenseServer:
            data.update({'licensingServers': licenseServer})
        if licenseMode:
            data.update({'mode': licenseMode})
        if licenseTier:
            data.update({'tier': licenseTier})

        response = self.patch(self.sessionUrl+'/globals/licensing', data=data)
        self.showLicenseDetails()

    def showLicenseDetails(self):
        """
        Description
           Display the new session's license details.

        Syntax
            GET: /api/v1/sessions/{id}/globals/licensing
        """
        response = self.get(self.sessionUrl+'/globals/licensing')
        self.logInfo('\nVerifying sessionId license server: %s' % self.sessionUrl, timestamp=False)
        self.logInfo('\t%s' % response.json()['licensingServers'], timestamp=False)
        self.logInfo('\t%s'%  response.json()['mode'], timestamp=False)
        self.logInfo('\t%s' % response.json()['tier'], timestamp=False)

    def getAllOpenSessionIds(self):
        """
        Description
           Get a list of open session IDs and some session metas.
           
        Syntax
            GETE: /api/v1/sessions

        Return
            A dict
            
        """
        response = self.get(self.sessionUrl)
        sessionIdDict = {}
        from pprint import pprint
        pprint(response.json())
        return
        for session in response.json():
            href = session['links'][0]['href']
            id = session['id']
            username = session['userName']
            state = session['state']
            createOn = session['createdOn']

            sessionIdDict[id] = {'id': id, 'sessionIdUrl': href, 'username': username, 'state': state, 'createdOn': createOn}

        return sessionIdDict

    def linuxServerStopAndDeleteSession(self):
        """
        Description
           Wrapper to stop and delete the session ID on the Linux API server.

        Requirements
           linuxServerStopOperations()
           linuxServerDeleteSession()

        Syntax
           GET = /api/v1/sessions/{id}
        """
        if self.serverOs == 'linux' and self.deleteSessionAfterTest==True:
            self.linuxServerStopOperations()
            self.linuxServerDeleteSession()

    def linuxServerStopOperations(self, sessionId=None):
        """
        Description
           Stop the session ID on the Linux API server.

        Parameter
           sessionId: (str): The session ID to stop.

        Requirement
           self.linuxServerWaitForSuccess()

        Syntax
            POST: /api/v1/sessions/{id}/operations/stop
        """
        if sessionId != None:
            sessionId = sessionId
        else:
            sessionId = self.sessionId

        response = self.post(sessionId+'/operations/stop')
        if self.linuxServerWaitForSuccess(response.json()['url'], timeout=90) == 1:
            raise IxNetRestApiException

    def linuxServerDeleteSession(self, sessionId=None):
        """
        Description
           Delete the session ID on the Linux API server.

        Paramter
          sessionId: (str): The session ID to delete on the Linux API server.

        Syntax
            DELETE: /api/v1/sessions/{id}/operations/stop
        """
        if sessionId != None:
            sessionId = sessionId
        else:
            sessionId = self.sessionId

        response = self.delete(sessionId)

    def linuxServerWaitForSuccess(self, url, timeout=120):
        """
        Description
           Wait for a success completion on the Linux API server.

        Paramters
           url: (str): The URL's ID of the operation to verify.
           timeout: (int): The timeout value.
        """
        data = {'applicationType': 'ixnrest'}
        jsonHeader = {'content-type': 'application/json', 'x-api-key': self.apiKey}
        self.logInfo('linuxServerWaitForSuccess')
        for counter in range(1,timeout+1):
            response = self.get(url, data=data, silentMode=True)
            currentStatus = response.json()['message']
            self.logInfo('\tCurrentStatus: {0}:  {1}/{2} seconds'.format(currentStatus, counter, timeout), timestamp=False)
            if counter < timeout+1 and currentStatus != 'Operation successfully completed':
                time.sleep(1)
            if counter == timeout+1 and currentStatus != 'Operation successfully completed':
                return 1
            if counter < timeout+1 and currentStatus == 'Operation successfully completed':
                return 0

    def newBlankConfig(self):
        """
        Description
           Start a new blank configuration.

        Requirement
            self.waitForComplete()

        Syntax:
           /api/v1/sessions/{1}/ixnetwork/operations/newconfig
        """
        url = self.sessionUrl+'/operations/newconfig'
        response = self.post(url)
        url = self.sessionUrl+'/operations/newconfig/'+response.json()['id']
        self.waitForComplete(response, url)

    def refreshHardware(self, chassisObj):
        """
        Description
           Refresh the chassis

        Parameter
           chassisObj: (str):The chassis object.
                           Ex: /api/v1/sessions/{1}/ixnetwork/availableHardware/chassis/1

        Requirement
           self.waitForComplete()

        Syntax
            /api/v1/sessions/{1}/ixnetwork/availableHardware/chassis/operations/refreshinfo
        """
        response = self.post(self.sessionUrl+'/availableHardware/chassis/operations/refreshinfo', data={'arg1': [chassisObj]})
        self.waitForComplete(response, self.sessionUrl+'/availableHardware/chassis/operations/refreshinfo')

    def query(self, data, silentMode=False):
        """
        Description
           Query for objects using filters.

        Paramater
           silentMode: (bool): True: Don't display any output on stdout.

        Notes
            Assuming this is a BGP configuration, which has two Topologies.
            Below demonstrates how to query the BGP host object by
            drilling down the Topology by its name and the specific the BGP attributes to modify at the
            BGPIpv4Peer node: flap, downtimeInSec, uptimeInSec.
            The from '/' is the entry point to the API tree.
            Notice all the node. This represents the API tree from the / entry point and starting at
            Topology level to the BGP host level.

        Notes
           Use the API Browser tool on the IxNetwork GUI to view the API tree.
            data: {'from': '/',
                    'nodes': [{'node': 'topology',    'properties': ['name'], 'where': [{'property': 'name', 'regex': 'Topo1'}]},
                              {'node': 'deviceGroup', 'properties': [], 'where': []},
                              {'node': 'ethernet',    'properties': [], 'where': []},
                              {'node': 'ipv4',        'properties': [], 'where': []},
                              {'node': 'bgpIpv4Peer', 'properties': ['flap', 'downtimeInSec', 'uptimeInSec'], 'where': []}]
                }

        Requirements
            self.waitForComplete()

        Examples
            response = restObj.query(data=queryData)
            bgpHostAttributes = response.json()['result'][0]['topology'][0]['deviceGroup'][0]['ethernet'][0]['ipv4'][0]['bgpIpv4Peer'][0]

            # GET THE BGP ATTRIBUTES TO MODIFY
            bgpHostFlapMultivalue = bgpHostAttributes['flap']
            bgpHostFlapUpTimeMultivalue = bgpHostAttributes['uptimeInSec']
            bgpHostFlapDownTimeMultivalue = bgpHostAttributes['downtimeInSec']

            restObj.configMultivalue(bgpHostFlapMultivalue, multivalueType='valueList', data={'values': ['true', 'true']})
            restObj.configMultivalue(bgpHostFlapUpTimeMultivalue, multivalueType='singleValue', data={'value': '60'})
            restObj.configMultivalue(bgpHostFlapDownTimeMultivalue, multivalueType='singleValue', data={'value': '30'})
        """
        url = self.sessionUrl+'/operations/query'
        reformattedData = {'selects': [data]}
        response = self.post(url, data=reformattedData, silentMode=silentMode)
        self.waitForComplete(response, url+'/'+response.json()['id'], silentMode=silentMode)
        return response

    def configMultivalue(self, multivalueUrl, multivalueType, data):
        """
        Description
           Configure multivalues.

        Parameters
           multivalueUrl: (str): The multivalue: /api/v1/sessions/{1}/ixnetwork/multivalue/1
           multivalueType: (str): counter|singleValue|valueList
           data: (dict): singleValue: data={'value': '1.1.1.1'})
                             valueList:   data needs to be in a [list]:  data={'values': [list]}
                             counter:     data={'start': value, 'direction': increment|decrement, 'step': value}
        """
        if multivalueType == 'counter':
            # Examples: macAddress = {'start': '00:01:01:00:00:01', 'direction': 'increment', 'step': '00:00:00:00:00:01'}
            #          data=macAddress)
            self.patch(self.httpHeader+multivalueUrl+'/counter', data=data)

        if multivalueType == 'singleValue':
            # data={'value': value}
            self.patch(self.httpHeader+multivalueUrl+'/singleValue', data=data)

        if multivalueType == 'valueList':
            # data={'values': ['item1', 'item2']}
            self.patch(self.httpHeader+multivalueUrl+'/valueList', data=data)

    def getMultivalueValues(self, multivalueObj, silentMode=False):
        """
        Description
           Get the multivalue values.

        Parameters
           multivalueObj: (str): The multivalue object: /api/v1/sessions/{1}/ixnetwork/multivalue/208
           silentMode: (bool): True=Display the GET and status code. False=Don't display.

        Requirements
           self.waitForComplete()
        """
        response = self.get(self.httpHeader+multivalueObj+'?includes=count', silentMode=silentMode)
        count = response.json()['count']
        if silentMode == False:
            self.logInfo('getMultivalueValues: {0} Count={1}'.format(multivalueObj, count))
        data = {'arg1': multivalueObj,
                'arg2': 0,
                'arg3': count
                }
        response = self.post(self.sessionUrl+'/multivalue/operations/getValues', data=data, silentMode=silentMode)
        self.waitForComplete(response, self.sessionUrl+'/operations/multivalue/getValues'+response.json()['id'])
        return response.json()['result']

    def getObjAttributeValue(self, obj, attribute):
        """
        Description
           Based on the object handle, get any property attribute and return the value.

        Parameter
           obj: <str:obj>: An object handle:
                For example: If you want the ethernet MTU, then pass in the ethernet object handle:
                            /api/v1/sessions/{id}/ixnetwork/topology/{id}/deviceGroup/{id}/ethernet/{id}
                            and set attribute='mtu'

        Note:
           Where to get the object's attribute names:
              - Use the API browser and go to your object.
              - All the attributes are listed on the right pane.
        """
        response = self.get(self.httpHeader + obj)

        # value: Could be /api/v1/sessions/{id}/ixnetwork/multivalue/{id} or the actual value
        value = response.json()[attribute]
        if type(value) == str and 'multivalue' in value:
            multivalueObj = value
            value = self.getMultivalueValues(multivalueObj)
            return value
        else:
            return value

    def stdoutRedirect(self):
        """
        Description
           For Robot Framework.  Robot captures the stdout. This stdoutRedirect
           will redirect the output back to stdout so you could see the test progress
           and to troubleshoot.
        """
        for attr in ('stdin', 'stdout', 'stderr'):
            setattr(sys, attr, getattr(sys, '__%s__' %attr))

    @staticmethod
    def prettyprintAllOperations(sessionUrl):
        """
        Description
           A staticmethod to rendering a nice output of an operations options and descriptions.

        Parameter
           sessionUrl: (str): http://{apiServerIp}:{port}/api/v1/sessions/1/ixnetwork

        Syntax:
            /api/v1/sessions/{1}/ixnetwork/operations
        """
        response = self._session.request('GET', sessionUrl+'/operations')
        for item in response.json():
            if 'operation' in item.keys():
                print('\n', item['operation'])
                print('\t%s' % item['description'])
                if 'args' in item.keys():
                    for nestedKey,nestedValue in item['args'][0].items():
                        print('\t\t%s: %s' % (nestedKey, nestedValue))

    @staticmethod
    def printDict(obj, nested_level=0, output=sys.stdout):
        """
        Description
           Print each dict key with indentions for human readability.
        """
        spacing = '   '
        spacing2 = ' '
        if type(obj) == dict:
            print( '%s' % ((nested_level) * spacing), file=output)
            for k, v in obj.items():
                if hasattr(v, '__iter__'):
                    print('%s%s:' % ( (nested_level+1) * spacing, k), file=output, end='')
                    IxNetRestMain.printDict(v, nested_level+1, output)
                else:
                    print('%s%s: %s' % ( (nested_level + 1) * spacing, k, v), file=output)

            print('%s' % (nested_level * spacing), file=output)
        elif type(obj) == list:
            print('%s[' % ((nested_level) * spacing), file=output)
            for v in obj:
                if hasattr(v, '__iter__'):
                    IxNetRestMain.printDict(v, nested_level + 1, file=output)
                else:
                    print('%s%s' % ((nested_level + 1) * spacing, v), file=output)
            print('%s]' % ((nested_level) * spacing), output)
        else:
            print('%s%s' % ((nested_level * spacing2), obj), file=output)

    def placeholder():
        pass





import re

from IxNetRestApiProtocol import Protocol

class Globals(object):
    def __init__(self, ixnObj=None):
        """
        Parameters
           ixnObj: <str>: The main connection object.
        """
        self.ixnObj = ixnObj
        self.protocolObj = Protocol(ixnObj)

    def dhcpV4ClientStartStopRate(self, endpoint='startRate', **kwargs):
        """
        Description
           Configure startRate|stopRate settings for DHCP V4 Client.

        Parameters
           endpoint: <str|object endpoint>: startRate|stopRate

           **kwargs: Any attribute for the /globals/topology/dhcpv4client/startRate|stopRate endpoint.
                     enabled = bool
                     interval = int
                     maxOutstanding = int
                     rate = int
                     rowNames = list
                     scalePortMode = str: port|deviceGroup
        Usage:
           globalObj.dhcpV4ClientStartStopRate(rate=500,
                                               maxOutstanding=600,
                                               enabled=True,
                                               interval=3000,
                                               scaleMode='port'
                                               )

        """
        restApi = '/globals/topology/dhcpv4client/{0}?links=true'.format(endpoint)

        response = self.ixnObj.get(self.ixnObj.sessionUrl + restApi)
        for key,value in response.json().items():
            if key != 'links':
                if bool(re.search('multivalue', str(value))) == True:
                    if key in kwargs:
                        multiValue = response.json()[key]
                        self.ixnObj.patch(self.ixnObj.httpHeader+multiValue+"/singleValue", data={'value': kwargs[key]})
                else:
                    if key in kwargs:
                        self.ixnObj.patch(self.ixnObj.sessionUrl + restApi, data={key: kwargs[key]})




import time
from .IxNetRestApi import IxNetRestApiException

class PortMgmt(object):
    def __init__(self, ixnObj=None):
        self.ixnObj = ixnObj

    def getSelfObject(self):
        # For Python Robot Framework support
        return self

    def setMainObject(self, mainObject):
        # For Python Robot Framework support
        self.ixnObj = mainObject

    def connectToVChassis(self, chassisIp):
        # Connects to the virtual chassis

        url = self.ixnObj.sessionUrl+'/operations/connecttochassis'
        data = {"arg1": chassisIp}

        response = self.ixnObj.post(url, data=data)
        if response == 1: return 1
        if response.json()['state'] == 'ERROR':
            self.ixnObj.logInfo('connectToVChassis error: %s' % response.json()['result'])
            return 1
        else:
            self.ixnObj.logInfo('connectToVChassis: Successfully connected to chassis: %s' % chassisIp)
            while response.json()["state"] == "IN_PROGRESS" or response.json()["state"] == "down":
                if timeout == 0:
                    break
                time.sleep(1)
                response = request.get(self.ixnObj.sessionUrl)
                state = response.json()["state"]
                self.ixnObj.logInfo("\t\t%s" % state)
                timeout = timeout - 1
            return 0

    def connectIxChassis(self, chassisIp, timeout=30, **kwargs):
        """
        Description
           Connect to an Ixia chassis.

        Parameter
           chassisIp: <str>|<list>: A string or a list of chassis IP addresses.
           timeout: <int>: Default=30 seconds. The amount of time to wait for the 
                           chassis to be in the ready state.

           kwargs: Any chassis attributes and values. For example, if two chassis' are dasisy chained, include:
                   chainTopology=None, masterChassis='10.10.10.1', sequenceId=1

        Syntax
           /api/v1/sessions/{id}/ixnetwork/availableHardware/chassis

        """
        if isinstance(chassisIp, list) == False:
            chassisIp = chassisIp.split(' ')

        chassisObjList = []
        url = self.ixnObj.sessionUrl+'/availableHardware/chassis'
        for chassisIpAddress in chassisIp:
            data = {'hostname': chassisIpAddress}
            if kwargs:
                data.update(kwargs)

            response = self.ixnObj.post(url, data=data)
            if type(response.json()) == list:
                # 8.50 json response is a list.
                chassisIdObj = response.json()[0]['links'][0]['href']
            else:
                chassisIdObj = response.json()['links'][0]['href']
 
            self.ixnObj.logInfo('\n', timestamp=False)
            # Chassis states: down, polling, ready
            for timer in range(1,timeout+1):
                response = self.ixnObj.get(self.ixnObj.httpHeader + chassisIdObj, silentMode=True)
                currentStatus = response.json()['state']
                self.ixnObj.logInfo('connectIxChassis {0}: Status: {1}. Wait {2}/{3} seconds'.format(
                    chassisIpAddress, currentStatus, timer, timeout), timestamp=False)
                if currentStatus != 'ready' and timer < timeout:
                    time.sleep(1)
                if currentStatus != 'ready' and timer == timeout:
                    raise IxNetRestApiException('connectIxChassis: Connecting to chassis {0} failed'.format(chassisIpAddress))
                if currentStatus == 'ready' and timer < timeout:
                    chassisObjList.append(chassisIdObj)
                    break

        # http://192.168.70.127:11009/api/v1/sessions/1/ixnetwork/availableHardware/chassis/1
        return chassisObjList

    
    def disconnectIxChassis(self, chassisIp):
        """
        Description
            Disconnect the chassis (both hardware or virtualChassis).

        Parameter
            chassisIp: <str>: The chassis IP address.

        Syntax
            /api/v1/sessions/{id}/ixnetwork/availableHardware/chassis/{id}
        """
        url = self.ixnObj.sessionUrl+'/availableHardware/chassis'
        response = self.ixnObj.get(url)
        for eachChassisId in response.json():
            if eachChassisId['hostname'] == chassisIp:
                chassisIdUrl = eachChassisId['links'][0]['href']
                self.ixnObj.logInfo('disconnectIxChassis: %s' % chassisIdUrl)
                response = self.ixnObj.delete(self.ixnObj.httpHeader+chassisIdUrl)

    def getChassisId(self, chassisIp):
        """
        Description
           Get the chassis ID based on the chassis IP address.
        
        Parameter
           chassisIp: <str>: The chassis IP address
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/availableHardware/chassis')
        for eachChassis in response.json():
            if eachChassis['ip'] == chassisIp:
                return eachChassis['id']

    def connectVportTo(self, portList):
        """
        Description
           This function assumes that a list of virtual ports are created.
           Connect the portList to the next vport that is not connected to any physical port.

        portList: <list>: A list of ports in a list: [[ixChassisIp, card, port]]
        """
        self.createVports(portList)
        vportObjectList = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        portListIndex = 0
        for vportObj in vportObjectList.json():
            print('\n', vportObj)
            connectedTo = vportObj['connectedTo']
            vportHref = vportObj['links'][0]['href']
            if connectedTo == 'null':
                # RW: 'connectedTo': '/api/v1/sessions/1/ixnetwork/availableHardware/chassis/1/card/1/port/1'
                # ReadOnly: 'assignedTo': '192.168.70.11:1:1'
                chassisIp = portList[portListIndex][0]
                cardNumber = portList[portListIndex][1]
                portNumber = portList[portListIndex][2]
                chassisId = self.getChassisId(chassisIp)
                self.ixnObj.patch(self.ixnObj.sessionUrl+'/availableHardware/chassis/'+str(chassisId)+'/card/'+str(cardNumber)+'/port/'+str(portNumber))
                data = '/api/v1/sessions/{0}/ixnetwork/availableHardware/chassis/{1}/card/{2}/port/{3}'.format(
                    self.ixnObj.sessionIdNumber, chassisId, cardNumber, portNumber)
                self.ixnObj.patch(self.ixnObj.httpHeader+vportHref, data={'connectedTo': data})
                if portListIndex < len(portList):
                    portListIndex += 1
                    continue
                else:
                    break

    def createVports(self, portList=None, rawTrafficVportStyle=False):
        """
        Description
           This API creates virtual ports based on a portList.
           Next step is to call assignPort.

        Parameters
            portList: <list>: Pass in a list of ports in the format of ixChassisIp, slotNumber, portNumber
                              portList = [[ixChassisIp, '1', '1'], [ixChassisIp, '2', '1']]

            rawTrafficVportStyle: <bool>: For raw Traffic Item src/dest endpoints, vports must be in format:
                               /api/v1/sessions1/vport/{id}/protocols

         Return
            A list of vports
        """
        createdVportList = []
        for index in range(0, len(portList)):
            self.ixnObj.logInfo('Creating a new virtual port')
            response = self.ixnObj.post(self.ixnObj.sessionUrl+'/vport')
            vportObj = response.json()['links'][0]['href']
            if rawTrafficVportStyle:
                createdVportList.append(vportObj+'/protocols')
            else:
                createdVportList.append(vportObj)
            if portList != None:
                response = self.ixnObj.get(self.ixnObj.httpHeader+vportObj)
                card = portList[index][1]
                port = portList[index][2]
                portNumber = str(card)+'/'+str(port)
                self.ixnObj.logInfo('\tName: %s' % portNumber)
                response = self.ixnObj.patch(self.ixnObj.httpHeader+vportObj, data={'name': portNumber})

        if createdVportList == []:
            raise IxNetRestApiException('No vports created')

        self.ixnObj.logInfo('createVports: %s' % createdVportList)
        return createdVportList

    def getVportObjectByName(self, portName):
        """
        Description:
           Get the vport object by the specified port name.

        Parameter
           portName: <str>: The name of the virtual port.
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = ["%s/vport/%s" % (self.ixnObj.sessionUrl, str(i["id"])) for i in response.json()]
        for vportObj in vportList:
            response = self.ixnObj.get(vportObj)
            if response.json()['name'] == portName:
                return vportObj
        return None

    def getVportName(self, vportObj):
        """
        Description
           Get the name of the vport by the specified vport object

        Parameter
            vportObj: <str>: /api/v1/sessions/1/ixnetwork/vport/1
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader+vportObj)
        return response.json()['name']

    def linkUpDown(self, port, action='down'):
        """
        Description
            Flap a port up or down.

        Parameters
            port: <list>: A list of ports in a list.  [[ixChassisIp, str(card), str(port)]] -> ['10.10.10.1', '1', '3']
            action: <str>: up|down
        """
        vport = self.ixnObj.getVports([port])[0]
        self.ixnObj.post(self.ixnObj.sessionUrl+'/vport/operations/linkUpDn', data={'arg1': [vport], 'arg2': action})

    def getAllVportList(self):
        """
        Description
            Returns a list of all the created virtual ports

        Returns
            List of vports: ['/api/v1/sessions/1/ixnetwork/vport/1', '/api/v1/sessions/1/ixnetwork/vport/2']
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = ['%s' % vport['links'][0]['href'] for vport in response.json()]
        return vportList

    def getVports(self, portList):
        """
        Description
            Get the vports for the portList

        Parameter
            portList: <list>: A list of ports in a list: [[str(chassisIp), str(slotNumber), str(portNumber)]]
                      Example 1: [[ixChassisIp, '1', '1']]
                      Example 2: [[ixChassisIp, '1', '1'], [ixChassisIp, '2', '1']]

        Returns
             Vports in a list: ['/api/v1/sessions/1/ixnetwork/vport/1', /api/v1/sessions/1/ixnetwork/vport/2']
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = []

        for vportAttributes in response.json():
            currentVportId = vportAttributes['links'][0]['href']
            # "assignedTo": "192.168.70.10:1:1
            assignedTo = vportAttributes['assignedTo']
            if assignedTo == '':
                continue

            chassisIp = assignedTo.split(':')[0]
            cardNum = assignedTo.split(':')[1]
            portNum = assignedTo.split(':')[2]
            port = [chassisIp, cardNum, portNum]

            if port in portList:
                # ['192.168.70.10', '1', '1']
                vport = vportAttributes['links'][0]['href']
                vportList.append(vport)

        # Returns:
        # ['/api/v1/sessions/1/ixnetwork/vport/1', /api/v1/sessions/1/ixnetwork/vport/2']
        return vportList

    def getPhysicalPortsFromCreatedVports(self):
        """
        Description
            Get all the ports that are configured.

        Return
            None or a list of ports in format: [['192.168.70.11', '1', '1'], ['192.168.70.11', '2', '1']]
        """
        portList = []
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = ['%s' % vport['links'][0]['href'] for vport in response.json()]

        for eachVport in vportList:
            response = self.ixnObj.get(self.ixnObj.httpHeader+eachVport)
            assignedTo = response.json()['assignedTo']
            # assignedTo: 192.168.70.11:2:1
            if assignedTo:
                chassis = assignedTo.split(':')[0]
                card = assignedTo.split(':')[1]
                port = assignedTo.split(':')[2]
                portList.append([chassis, card, port])
        return portList

    def getPhysicalPortFromVport(self, vportList):
        """
        Description
            Get the physical ports assigned to the vport objects.

        Parameter
            vportList: ['/api/v1/sessions/1/ixnetwork/vport/1']

        Returns
            A list of ports: ['192.168.70.11:1:1']
        """
        portList = []
        for eachVport in vportList:
            response = self.ixnObj.get(self.ixnObj.httpHeader+eachVport)
            assignedTo = response.json()['assignedTo']
            if assignedTo:
                portList.append(assignedTo)
        return portList

    def verifyPortConnectionStatus(self, vport=None):
        """
        Description
           Verify port connection status for errors such as License Failed, 
           Version Mismatch, Incompatible IxOS version, or any other error.
        """
        self.ixnObj.logInfo('verifyPortConnectionStatus')
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')

        for vport in response.json():
            connectionStatus = vport['connectionStatus']
            if 'Port Released' in connectionStatus:
                raise IxNetRestApiException(connectionStatus)

    def assignPorts(self, portList, forceTakePortOwnership=True, createVports=False,
                    rawTraffic=False, configPortName=True, timeout=120):
        """
        Description
            Assuming that you already connected to an ixia chassis and ports are available for usage.
            Use this API to assign physical ports to the virtual ports.

        Parameters
            portList: <list>: A list of ports in a list: [ [ixChassisIp, '1','1'], [ixChassisIp, '1','2'] ]
            forceTakePortOwnership: <bool>: True = Forcefully take ownership of portList.

            createVports: <bool>: Optional:
                          If True: Create vports to the amount of portList.
                          If False: Automatically create vport on the server side. Optimized for port bootup performance. 

            rawTraffic: <bool>:  If traffic item is raw, then vport needs to be /vport/{id}/protocols
            resetPortCput: <bool>: Default=False. Some cards like the Novus 10GigLan requires a cpu reboot.
            timeout: <int>: Timeout for port up state. Default=90 seconds.

        Syntaxes
            POST: /api/v1/sessions/{id}/ixnetwork/operations/assignports
                  data={arg1: [{arg1: ixChassisIp, arg2: 1, arg3: 1}, {arg1: ixChassisIp, arg2: 1, arg3: 2}],
                        arg2: [],
                        arg3: ['/api/v1/sessions/{1}/ixnetwork/vport/1',
                               '/api/v1/sessions/{1}/ixnetwork/vport/2'],
                        arg4: true}  <-- True will clear port ownership
                  headers={'content-type': 'application/json'}

            GET:  /api/v1/sessions/{id}/ixnetwork/operations/assignports/1
                  data={}
                  headers={}
            Expecting:   RESPONSE:  SUCCESS
        """
        # Verify if the portList has duplicates.
        self.verifyForDuplicatePorts(portList)

        # Verify if there is existing vports. If yes, user either loaded a saved config file or 
        # the configuration already has vports.
        # If loading a saved config file and reassigning ports, assign ports to existing vports.
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')

        # If response.json() != [], means there are existing vports created already.
        if response.json() != []:
            mode = 'modify'
            preamble = self.ixnObj.sessionUrl.split('/api')[1]
            vportList = ["/api%s/vport/%s" % (preamble, str(i["id"])) for i in response.json()]
            if len(vportList) != len(portList):
                raise IxNetRestApiException('assignPorts: The amount of configured virtual ports:{0} is not equal to the amount of  portList:{1}'.format(len(vportList), len(portList)))

        else:
            if createVports == False:
                vportList = []

            if createVports:
                self.createVports(portList)
                response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
                preamble = self.ixnObj.sessionUrl.split('/api')[1]

                vportList = ["/api%s/vport/%s" % (preamble, str(i["id"])) for i in response.json()]
                if len(vportList) != len(portList):
                    raise IxNetRestApiException('assignPorts: The amount of configured virtual ports:{0} is not equal to the amount of  portList:{1}'.format(len(vportList), len(portList)))

        #data = {"arg1": [], "arg2": [], "arg3": vportList, "arg4": 'true'}
        data = {"arg1": [], "arg2": [], "arg3": vportList, "arg4": forceTakePortOwnership}
        [data["arg1"].append({"arg1":str(chassis), "arg2":str(card), "arg3":str(port)}) for chassis,card,port in portList]
        url = self.ixnObj.sessionUrl+'/operations/assignports'
        response = self.ixnObj.post(url, data=data)
        response = self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'], silentMode=False, timeout=timeout, ignoreException=True)

        if response.json()['state'] == 'EXCEPTION':
            # Some cards like the Novus 10gLan sometimes requires a cpu reboot.
            # To reboot the port cpu, the ports have to be assigned to a vport first.
            # So it has to be done at this spot.
            self.resetPortCpu(vportList=vportList, portList=portList)
            self.verifyPortState()
            
            raise IxNetRestApiException('assignPort Error: {}'.format(response.json()['message']))

        elif response.json()['state'] == 'IN_PROGRESS':
            raise IxNetRestApiException('assignPort Error: Port failed to boot up after 120 seconds')

        else:
            response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
            for vport in response.json():
                chassisIp = vport['assignedTo'].split(':')[0]
                slot = vport['assignedTo'].split(':')[1]
                port = vport['assignedTo'].split(':')[2]
                currentPort = [chassisIp, int(slot), int(port)]
                for chassis,card,port in portList:
                    currentPortList = [chassis, int(card), int(port)]
                    if set(currentPort) & set(currentPortList):
                        if 'License Failed' in vport['connectionStatus']:
                            raise IxNetRestApiException('Port License failed.')
                        if vport['connectionStatus'] == 'connectedLinkDown':
                            raise IxNetRestApiException('Port link connection is down: {0}'.format(vport['assignedTo']))

        if configPortName:
            # Name the vports
            for vportObj in self.getAllVportList():
                port = self.getPhysicalPortFromVport([vportObj])[0]
                chassisIp = port.split(':')[0]
                card = port.split(':')[1]
                port = port.split(':')[2]
                self.ixnObj.patch(self.ixnObj.httpHeader+vportObj, data={'name': 'Port'+card+'_'+port})

        if rawTraffic:
            vportProtocolList = []
            for vport in self.getAllVportList():
                vportProtocolList.append(vport+'/protocols')
            return vportProtocolList
        else:
            return vportList

    def unassignPorts(self, deleteVirtualPorts=False):
        """
        Description
            Unassign all virtual ports from the configuration.

        Parameters
            deleteVirtualPorts: <bool>:
                                True = Delete the virtual ports from the configuration.
                                False = Unassign the virtual ports from the configuration.

        Syntaxes
            POST:  http://{apiServerIp:port}/api/v1/sessions/{id}/ixnetwork/vport/operations/unassignports
                   data={arg1: [http://{apiServerIp:port}/api/v1/sessions/{id}/ixnetwork/vport/1,
                                http://{apiServerIp:port}/api/v1/sessions/{id}/ixnetwork/vport/2],
                         arg2: true|false}
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = ["%s/vport/%s" % (self.ixnObj.sessionUrl, str(i["id"])) for i in response.json()]
        url = self.ixnObj.sessionUrl+'/vport/operations/unassignports'
        response = self.ixnObj.post(url, data={'arg1': vportList, 'arg2': deleteVirtualPorts})
        self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/vport/operations/unassignports/'+response.json()['id'], timeout=120)

    def releaseAllPorts(self):
        """
        Description
            Release all the connected ports.
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = ["%s/vport/%s" % (self.ixnObj.sessionUrl, str(i["id"])) for i in response.json()]
        url = self.ixnObj.sessionUrl+'/vport/operations/releaseport'
        response = self.ixnObj.post(url, data={'arg1': vportList})
        if response.json()['state'] == 'SUCCESS': return 0
        if response.json()['id'] != '':
            self.ixnObj.waitForComplete(response, url+'/'+response.json()['id'], timeout=120)
            
    def releasePorts(self, portList):
        """
        Description
            Release the specified ports in a list.

        Parameter
            portList: <list>: A list of ports in a list, to release in format of...
                      [[ixChassisIp, str(cardNum), str(portNum)], [], [] ...]
        """
        vportList = []
        for port in portList:
            vport = self.getVports([port])
            if vport == []:
                continue
            vportList.append(vport[0])

        url = self.ixnObj.sessionUrl+'/vport/operations/releaseport'
        response = self.ixnObj.post(url, data={'arg1': vportList})
        self.ixnObj.waitForComplete(response, url+'/'+response.json()['id'], timeout=120)

    def resetPortCpu(self, vportList=None, portList=None, timeout=90):
        """
        Description
            Reset/Reboot ports CPU.
            Must call IxNetRestApi.py waitForComplete() afterwards to verify port state
        
        Parameter
            vportList: <list>: A list of one or more vports to reset.
        """
        url = self.ixnObj.sessionUrl+'/vport/operations/resetportcpu'
        if vportList == None:
            vportList = self.getVports(portList)

        response = self.ixnObj.post(url, data={'arg1': vportList})
        self.ixnObj.waitForComplete(response, url + '/' +response.json()['id'], silentMode=False, timeout=timeout)


    def resetPortCpuAndFactoryDefault(self, vportList=None, portList=None, timeout=90):
        """
        Description
            Reset/Reboot ports CPU.and set to factory default
            Must call IxNetRestApi.py waitForComplete() afterwards to verify port state
        
        Parameter
            vportList: <list>: A list of one or more vports to reset.
        """
        url = self.ixnObj.sessionUrl+'/vport/operations/resetportcpuandfactorydefault'
        if vportList == None:
            vportList = self.getVports(portList)

        response = self.ixnObj.post(url, data={'arg1': vportList})
        self.ixnObj.waitForComplete(response, url + '/' +response.json()['id'], silentMode=False, timeout=timeout)

    def clearPortOwnership(self, portList):
        """
            Description
                Clear port ownership on the portList

            Parameters
                portList: <list>: A list of ports in a list: [[chassisIp, cardId, portId]]
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/availableHardware/chassis')
        for eachChassis in response.json():
            chassisIp = eachChassis['ip']
            chassisHref = eachChassis['links'][0]['href']

            for userPort in portList:
                userChassisIp = userPort[0]
                if userChassisIp != chassisIp:
                    continue
                userCardId = userPort[1]
                userPortId = userPort[2]
                url = self.ixnObj.httpHeader+chassisHref+'/card/'+str(userCardId)+'/port/'+str(userPortId)+'/operations/clearownership'
                data = {'arg1': [chassisHref+'/card/'+str(userCardId)+'/port/'+str(userPortId)]}
                self.ixnObj.post(url, data=data)

    def isPortConnected(self, portList):
        """
        Description
            Verify if the port is connected or released

        Parameters
            portList: <list>: A list of ports in a list:  [[ixChassisIp, str(cardNumber), str(portNumber)]]

        Return
            A list of 'connected' and 'released'.
        """
        returnValues = []
        for port in portList:
            vport = self.getVports([port])
            if vport == []:
                returnValues.append('released')
                continue
            response = self.ixnObj.get(self.ixnObj.httpHeader+vport[0])
            connectedStatus = response.json()['connectionStatus']
            print('\nisPortConnected:', port)
            if connectedStatus == 'Port Released':
                self.ixnObj.logInfo('\tFalse: %s' % connectedStatus)
                returnValues.append('released')
            else:
                self.ixnObj.logInfo('\tTrue: %s' % connectedStatus)
                returnValues.append('connected')
        return returnValues

    def verifyForDuplicatePorts(self, portList):
        """
        Description
           Verify if the portList has any duplicate ports.
           Raise an exception if true.
        """
        duplicatePorts = [x for n, x in enumerate(portList) if x in portList[:n]]
        if duplicatePorts:
            raise IxNetRestApiException('\nYour portList has duplicate ports {0}'.format(duplicatePorts))

    def arePortsAvailable(self, portList, raiseException=True):
        """
        Description:
           Verify if any of the portList is owned.

        Parameter: <list>: A list of ports in a list.
                   portList: [ ['192.168.70.11', '1', '1'], ['192.168.70.11', '2', '1'] ]

        raiseException: <bool>: To continue or not to continue if there is an error.

        Return:
            - List of ports that are currently owned
            - 0: If portList are available
        """
        # Verify if the portList has duplicates.
        self.verifyForDuplicatePorts(portList)

        self.ixnObj.logInfo('Verify if ports are currently owned')
        portOwnedList = []
        for port in portList:
            chassisIp = port[0]
            cardId = port[1]
            portId = port[2]
            try:
                queryData = {"from": "/availableHardware",
                                "nodes": [{"node": "chassis", "properties": ["ip"], "where": [{"property": "ip", "regex": chassisIp}]},
                                        {"node": "card", "properties": ["cardId"], "where": [{"property": "cardId", "regex": cardId}]},
                                        {"node": "port", "properties": ["portId", "owner"], "where": [{"property": "portId", "regex": portId}]}]}

                self.ixnObj.logInfo('Querying for %s/%s/%s' % (chassisIp, cardId, portId))
                queryResponse = self.ixnObj.query(data=queryData, silentMode=False)
                queryResponse.json()['result'][0]['chassis'][0]['ip']
                queryResponse.json()['result'][0]['chassis'][0]['card'][0]['id']
                queryResponse.json()['result'][0]['chassis'][0]['card'][0]['port'][0]['portId']
            except:
                raise IxNetRestApiException('\nNot found: {0}:{1}:{2}'.format(chassisIp, cardId, portId))

            self.ixnObj.logInfo('Port currently owned by: %s' % queryResponse.json()['result'][0]['chassis'][0]['card'][0]['port'][0]['owner'])
            if queryResponse.json()['result'][0]['chassis'][0]['card'][0]['port'][0]['owner'] != '':
                self.ixnObj.logInfo('Port is still owned: {0}/cardId:{1}/portId:{2}'.format(chassisIp, cardId, portId))
                portOwnedList.append([chassisIp, cardId, portId])

        self.ixnObj.logInfo('Ports are still owned: %s' % portOwnedList)

        if portOwnedList != []:
            if raiseException:
                raise IxNetRestApiException('arePortsAvailable: Ports are still owned')
            else:
                return portOwnedList
        return 0

    def verifyPortState(self, timeout=70):
        """
        Description
            Verify port states for all the vports connected to physical ports.

        Parameter
            timeout: <int>: The timeout value to declare as failed. Default=70 seconds.
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = [metaDatas["links"][0]['href'] for metaDatas in response.json()]
        for eachVport in vportList:
            for counter in range(1,timeout+1):
                stateResponse = self.ixnObj.get(self.ixnObj.httpHeader+eachVport+'?includes=state,connectionStatus', silentMode=True)
                assignedToResponse = self.ixnObj.get(self.ixnObj.httpHeader+eachVport+'?includes=assignedTo', silentMode=True)

                if 'Port Released' in stateResponse.json()['connectionStatus']:
                    raise IxNetRestApiException(stateResponse.json()['connectionStatus'])

                if stateResponse.json()['state'] == 'unassigned':
                    self.ixnObj.logWarning('\nThe vport {0} is not assigned to a physical port. Skipping this vport verification.'.format(eachVport))
                    break

                self.ixnObj.logInfo('Port: %s' % assignedToResponse.json()['assignedTo'])
                self.ixnObj.logInfo('\tVerifyPortState: %s\n\tWaiting %s/%s seconds' % (stateResponse.json()['state'], counter, timeout), timestamp=False)
                if counter < timeout and stateResponse.json()['state'] in ['down', 'busy']:
                    time.sleep(1)
                    continue
                if counter < timeout and stateResponse.json()['state'] in ['up', 'connectedLinkUp']:
                    break
                if counter == timeout and stateResponse.json()['state'] == 'down':
                    # Failed
                    raise IxNetRestApiException('Port failed to come up')

    def getVportFromPortList(self, portList):
        """
        Description
           Get a list of vports from the specified portList.

        Parameter
           portList: <list>: Format: [[ixChassisIp, cardNumber1, portNumber1], [ixChassisIp, cardNumber1, portNumber2]]
    
        Return
           A list of vports.
           [] if vportList is empty.
        """
        vportList = []
        for eachPort in portList:
            chassisIp = eachPort[0]
            card = eachPort[1]
            portNum = eachPort[2]
            port = chassisIp+':'+card+':'+portNum
            # {'href': '/api/v1/sessions/1/ixnetwork/',
            # 'vport': [{'id': 2, 'href': '/api/v1/sessions/1/ixnetwork/vport/2', 'assignedTo': '10.10.10.8:1:2'}]}
            queryData = {"from": "/",
                         "nodes": [{"node": "vport", "properties": ["assignedTo"],
                                    "where": [{"property": "assignedTo", "regex": port}]
                                }]}
            
            queryResponse = self.ixnObj.query(data=queryData, silentMode=False)
            vport = queryResponse.json()['result'][0]['vport']
            if vport == []:
                raise IxNetRestApiException('getVportFromPortList error: The port has no vport and not assigned. Check for port typo: {0}'.format(port))

            if vport:
                # Appending vportList: ['/api/v1/sessions/1/ixnetwork/vport/1', '/api/v1/sessions/1/ixnetwork/vport/2']
                vportList.append(vport[0]['href'])
        return vportList
                                
    def modifyPortMediaType(self, portList='all', mediaType='fiber'):
        """
        Description
           Modify the port media type: fiber, copper, SGMII

        Parameters
           portList: <'all'|list of ports>: 
                     <list>: Format: [[ixChassisIp, str(cardNumber1), str(portNumber1])]...]
                     Or if portList ='all', will modify all assigned ports to the specified mediaType.

           mediaType: <str>: copper, fiber or SGMII
        """
        self.ixnObj.logInfo('modifyPortMediaType: {0}'.format(mediaType))
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        if portList == 'all':
            #vportList = self.getAllVportList()
            portList = self.getPhysicalPortsFromCreatedVports()

        # vportList: ['/api/v1/sessions/1/ixnetwork/vport/1', '/api/v1/sessions/1/ixnetwork/vport/2']
        vportList = self.getVportFromPortList(portList)
        for vport in vportList:
            response = self.ixnObj.get(self.ixnObj.httpHeader+vport, silentMode=True)
            portType = response.json()['type']
            self.ixnObj.patch(self.ixnObj.httpHeader+vport+'/l1Config/'+portType, data={'media': mediaType})

    def configLoopbackPort(self, portList='all', enabled=True):
        """
        Description
           Configure port to loopback.

        Parameters
           portList: <'all'|list of ports>: 
                     <list>: Format: [[ixChassisIp, str(cardNumber1), str(portNumber1])]...]
                     Or if portList ='all', will modify all assigned ports to the specified mediaType.

           enabled: <bool>: True=Enable port to loopback mode.
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        if portList == 'all':
            portList = self.getPhysicalPortsFromCreatedVports()

        vportList = self.getVportFromPortList(portList)
        for vport in vportList:
            response = self.ixnObj.get(self.ixnObj.httpHeader+vport+'/l1Config', silentMode=False)
            currentPortType = response.json()['currentType']
            self.ixnObj.patch(self.ixnObj.httpHeader+vport+'/l1Config/'+currentPortType, data={'loopback': enabled})

    def setTxMode(self, vportList='all', txMode='interleaved', timeout=70):
        """
        Description
           set TxMode of the vports

        Parameter
           vportList: <list>: vports to set the transmitMode on.  Default = all
           txMode:    <str>: transmit mode setting -  can be either 'interleaved' or 'sequential'
           timeout:   <int>: the timeout value to declare as failed. Default=70 seconds.
        """
        if vportList == 'all':
            response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
            vportList = ['%s' % vport['links'][0]['href'] for vport in response.json()]

        # vportList: ['/api/v1/sessions/1/ixnetwork/vport/1', '/api/v1/sessions/1/ixnetwork/vport/2']
        for eachVport in vportList:
            self.ixnObj.patch(self.ixnObj.httpHeader+eachVport, data={'txMode':txMode})

    def configUdsRxFilters(self, portList='all', filterPalette=None, udsNum='1', udsArgs=None):
        """
        Description
           Configure rxFilters and User Defined Stats on a port

        Parameters
           portList: <'all'|list of ports>: 
                     <list>: Format: [[ixChassisIp, str(cardNumber1), str(portNumber1])]...]
                     Or if portList ='all', will modify all assigned ports to the specified mediaType.

           filterPalette: Filter Palette kwargs.
           udsNum: <string>:  uds number          
           udsArgs: uds kwargs.

        USAGE EXAMPLE:
           portMgmtObj.configUdsRxFilters(portList=[['10.113.9.219', '6', '1']], 
                                          filterPalette={'pattern1':'01', 'pattern1Mask':'FC',
                                                         'pattern1Offset':'15', 'pattern1OffsetType':'fromStartOfFrame', 
                                          udsNum=1
                                          udsArgs={'isEnabled':'true', 'patternSelector':'pattern1'})

           portMgmtObj.configUdsRxFilters(portList=[['10.113.9.219', '6', '1']], 
                                          filterPalette={'pattern2':'03', 'pattern2Mask':'FC',
                                                         'pattern2Offset':'19', 'pattern2OffsetType':'fromStartOfFrame', 
                                          udsNum=2
                                          udsArgs={'isEnabled':'true', 'patternSelector':'pattern2'})

        """
        self.ixnObj.logInfo('configUdsRxFilters: filterPalette={0}'.format(filterPalette))
        self.ixnObj.logInfo('\t\t uds={0} udsArgs={1}'.format(udsNum, udsArgs))
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        if portList == 'all':
            #vportList = self.getAllVportList()
            portList = self.getPhysicalPortsFromCreatedVports()

        # vportList: ['/api/v1/sessions/1/ixnetwork/vport/1', '/api/v1/sessions/1/ixnetwork/vport/2']
        vportList = self.getVportFromPortList(portList)
        for vport in vportList:
            response = self.ixnObj.get(self.ixnObj.httpHeader+vport, silentMode=True)
            portType = response.json()['type']
            self.ixnObj.patch(self.ixnObj.httpHeader+vport+'/l1Config/rxFilters/filterPalette', data=filterPalette)
            self.ixnObj.patch(self.ixnObj.httpHeader+vport+'/l1Config/rxFilters/uds/'+udsNum, data=udsArgs)




import sys, re 
from .IxNetRestApiFileMgmt import FileMgmt
from .IxNetRestApiPortMgmt import PortMgmt
from .IxNetRestApi import IxNetRestApiException

class PacketCapture(object):
    """
    ixnObj: The main connection object
    portObj: The port object

    """
    def __init__(self, ixnObj=None, portMgmtObj=None):
        """
        Parameters
            ixnObj: The main ixnObj object
            portMgmtObj: (Optional):  The PortMgmt object.
                         If you already have the portMgmt object, then pass it in. 
                         Otherwise, leave portMgmtObj=None

        Example:
            trafficObj = Traffic(mainObj)
            portMgmtObj = PortMgmt(mainObj)
            captureObj = PacketCapture(mainObj, portMgmtObj)
            captureObj.packetCaptureConfigPortMode(port=[ixChassisIp, '2', '1'], enableControlPlane=False, enableDataPlane=True)
            trafficObj.startTraffic()  <-- Make sure traffic is running in continuous mode
            captureObj.packetCaptureStart()
            time.sleep(10)
            captureObj.packetCaptureStop()
            trafficObj.stopTraffic()

            pktCaptureObj.getCapFile(port=[ixChassisIp, '2', '1'], typeOfCapture='data', saveToTempLocation='c:\\Results',
                                     localLinuxLocation='.', appendToSavedCapturedFile=None)

            pktCaptureObj.packetCaptureGetCurrentPackets(getUpToPacketNumber=5, capturePacketsToFile=True,
                                                         getSavedCapturedDotCapFile=True)

            captureObj.packetCaptureClearTabs()
        """
        self.ixnObj = ixnObj
        if portMgmtObj:
            self.portMgmtObj = portMgmtObj
        else:
            self.portMgmtObj = PortMgmt(ixnObj)

        self.enableControlPlane = False
        self.enableDataPlane = False

    def setMainObject(self, mainObject):
        # For Python Robot Framework support
        self.ixnObj = mainObject
        self.portMgmtObj.setMainObject(mainObject)

    def packetCaptureConfigPortMode(self, port, portRxMode='capture', enableControlPlane=True, enableDataPlane=True):
        """
        Description
           Enable|Disable port capturing for control plane and data plane.

           Values are true or false
              -softwareEnabled == Control Plane
              -hardwareEnabled == Data Plane

        Parameters
            port: <list>:  [ixChassisIp, '1', '3'] => [ixChasssisIp, str(cardNumber), str(portNumber)]
            portRxMode: <str>: capture|captureAndMeasure
            enableControlPlane: <bool>
            enableDataPlane: <bool>
        """
        if enableControlPlane == True:
            self.enableControlPlane = True
        if enableDataPlane == True:
            self.enableDataPlane = True

        self.captureRxPort = port
        vport = self.portMgmtObj.getVports([port])[0]
        self.ixnObj.patch(self.ixnObj.httpHeader+vport, data={'rxMode': portRxMode})
        self.ixnObj.patch(self.ixnObj.httpHeader+vport+'/capture', data={'softwareEnabled': enableControlPlane,
                                                                         'hardwareEnabled': enableDataPlane})

    def packetCaptureStart(self):
        """
        Start packet capturing
        """
        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/operations/startcapture')
        self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/operations/startcapture/'+response.json()['id'])

    def packetCaptureStop(self):
        """
        Stop packet capturing
        """
        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/operations/stopcapture')
        self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/operations/stopcapture/'+response.json()['id'])

    def packetCaptureClearTabs(self):
        """
        Remove all captured tabs on Windows IxNetwork GUI
        """
        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/operations/closeAllTabs')
        self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/operations/closeAllTabs/'+response.json()['id'])

    def packetCaptureGetCurrentPackets(self, getUpToPacketNumber=20, capturePacketsToFile=True):
        """
        Description
           Packet Capturing in wireshark style details. By default, it saved 7209 packet counts.
           It takes a long time to save all the packet header details into a file. 
           This API will default saving 20 packet counts. You could increase the packet count.

        Parameters
            getUpToPacketNumber: None|The last packet number to get. 
                                 Always starts at 1. If you state 10, then this function will get 1-10 packets.
            capturePacketsToFile: True|False
        """
        import subprocess, time

        if capturePacketsToFile:
            timestamp = int(time.time())
            if self.enableDataPlane:
                packetCaptureFilenameData = 'packetCaptureForData_'+str(timestamp)
                subprocess.call(['touch', packetCaptureFilenameData])

            if self.enableControlPlane:
                packetCaptureFilenameControl = 'packetCaptureForControl_'+str(timestamp)
                subprocess.call(['touch', packetCaptureFilenameControl])

            if self.enableDataPlane == False and self.enableControlPlane == False:
                raise IxNetRestApiException('\nPacketCapture Error: You must enable one of the options: enableDataPlane|enableControlPlane')
        
        vport = self.portMgmtObj.getVports([self.captureRxPort])[0]
        response = self.ixnObj.get(self.ixnObj.httpHeader+vport+'/capture')
        totalDataCapturedPackets = response.json()['dataPacketCounter']
        totalControlCapturedPackets = response.json()['controlPacketCounter']

        if type(totalDataCapturedPackets) != int:
            totalDataCapturedPackets = 0
        else:
            if getUpToPacketNumber != None:
                totalDataCapturedPackets = getUpToPacketNumber

        if type(totalControlCapturedPackets) != int:
            totalControlCapturedPackets = 0
        else:
            if getUpToPacketNumber != None:
                totalControlCapturedPackets = getUpToPacketNumber

        for eachTypeOfCaptures, totalCapturedPackets in zip(('data', 'control'), (totalDataCapturedPackets, totalControlCapturedPackets)):
            self.ixnObj.logInfo('Getting captured packets for capture type: {0}'.format(eachTypeOfCaptures))

            if capturePacketsToFile and int(totalCapturedPackets) != 0:
                timestamp = int(time.time())
                if self.enableDataPlane:
                    packetCaptureFilenameData = 'packetCaptureForData_'+str(timestamp)
                    subprocess.call(['touch', packetCaptureFilenameData])

                if self.enableControlPlane:
                    packetCaptureFilenameControl = 'packetCaptureForControl_'+str(timestamp)
                    subprocess.call(['touch', packetCaptureFilenameControl])

            for packetIndex in range(1, int(totalCapturedPackets)):
                self.ixnObj.logInfo('Getting captured packet index number: {}/{}'.format(packetIndex, getUpToPacketNumber))

                if self.enableDataPlane and eachTypeOfCaptures == 'data':
                    data = {'arg1': vport+'/capture/currentPacket', 'arg2': packetIndex}
                    response = self.ixnObj.post(self.ixnObj.sessionUrl+'/vport/capture/currentPacket/operations/getpacketfromdatacapture',
                                                data=data, silentMode=False)
                    self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/vport/capture/currentPacket/operations/getpacketfromdatacapture/'+response.json()['id'])

                if self.enableControlPlane and eachTypeOfCaptures == 'control':
                    data = {'arg1': vport+'/capture/currentPacket', 'arg2': packetIndex}
                    response = self.ixnObj.post(self.ixnObj.sessionUrl+'/vport/capture/currentPacket/operations/getpacketfromcontrolcapture',
                                                data=data, silentMode=False)

                    self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/vport/capture/currentPacket/operations/getpacketfromcontrolcapture/'+response.json()['id'])

                response = self.ixnObj.get(self.ixnObj.httpHeader+vport+'/capture/currentPacket/stack', silentMode=False)

                for eachStack in response.json():
                    displayName = eachStack['displayName']
                    stackIdObject = eachStack['links'][0]['href']
                    self.ixnObj.logInfo('\nStack: %s' % displayName)
                    if capturePacketsToFile:
                        if eachTypeOfCaptures == 'data':
                            with open(packetCaptureFilenameData, 'a') as packetCaptureFile:
                                packetCaptureFile.write('\nStack: %s\n' % displayName)

                        if eachTypeOfCaptures == 'control':
                            with open(packetCaptureFilenameControl, 'a') as packetCaptureFile:
                                packetCaptureFile.write('\nStack: %s\n' % displayName)

                    response = self.ixnObj.get(self.ixnObj.httpHeader+stackIdObject+'/field', silentMode=False)
                    for eachField in response.json():
                        fieldId = eachField['id']
                        fieldName = eachField['displayName']
                        fieldValue = eachField['fieldValue']
                        self.ixnObj.logInfo('\t{0}: {1}: {2}'.format(fieldId, fieldName, fieldValue))

                        if capturePacketsToFile:
                            if eachTypeOfCaptures == 'data':
                                with open(packetCaptureFilenameData, 'a') as packetCaptureFile:
                                    packetCaptureFile.write('\t{0}: {1}: {2}\n'.format(fieldId, fieldName, fieldValue))

                            if eachTypeOfCaptures == 'control':
                                with open(packetCaptureFilenameControl, 'a') as packetCaptureFile:
                                    packetCaptureFile.write('\t{0}: {1}: {2}\n'.format(fieldId, fieldName, fieldValue))

    def getCapFile(self, port, typeOfCapture='data', saveToTempLocation='c:\\Temp', localLinuxLocation='.', appendToSavedCapturedFile=None):
        """
        Get the latest captured .cap file from ReST API server to local Linux drive.

        Parameters
            port: Format:[IxiaIpAddress, slotNumber, cardNumber]
                  Example: [ixChassisIp, '2', '1']

            typeOfCapture: data|control

            saveToTempLocation: For Windows:
                                    Where to temporary save the .cap file on the ReST API server: 
                                    Provide any path with double backslashes: C:\\Temp.
                                    The folder will be created if it doesn't exists.

                                For Linux, value= 'linux'.
 
            localLinuxLocation: Where to save the .cap file on the local Linux machine.

            appendToSavedCapturedFile: Add a text string to the captured file.

        Example:
            captureObj.getCapFile([ixChassisIp, '2', '1'], 'control', 'c:\\Temp', '/home/hgee/test')

        Syntaxes:
               PATCH: /vport/2
               DATA: {'rxMode': 'captureAndMeasure'}

               PATCH: /vport/2/capture
               DATA: {'softwareEnabled': False, 'hardwareEnabled': True}
 
               # Set traffic item to send continuous packets
               PATCH: /traffic/trafficItem/1/configElement/<id>/transmissionControl
               DATA: {'type': 'continuous'}

               POST: /traffic/trafficItem/operations/generate
               DATA: {"arg1": ["/api/v1/sessions/<id>/ixnetwork/traffic/trafficItem/1"]}

               POST: /traffic/operations/apply
               DATA: {"arg1": "/api/v1/sessions/<id>/ixnetwork/traffic"}

               POST: /traffic/operations/start
               DATA: {"arg1": "https://192.168.70.12/api/v1/sessions/<id>/ixnetwork/traffic"}

               Start continuous traffic

               POST: /ixnetwork/operations/startcapture
               POST: /ixnetwork/operations/stopcapture
               POST: /ixnetwork/operations/savecapturefiles
               DATA: {"arg1": "packetCaptureFolder"}  <-- This could be any name.  Just a temporary folder to store the captured file.

               Wait for the /operations/savecapturefiles/<id> to complete.  May take up to a minute or more.

               For Windows API server:
                  POST: /ixnetwork/operations/copyfile
                  DATA: {"arg1": "c:\\Results\\port2_HW.cap", "arg2": "/api/v1/sessions/1/ixnetwork/files/port2_HW.cap"}
                  GET: /ixnetwork/files?filename=port2_HW.cap

               For Linux API server:
                  POST: /ixnetwork/operations/copyfile
                  DATA: {"arg1": "captures/packetCaptureFolder/port2_HW.cap", "arg2": "/api/v1/sessions/<id>/ixnetwork/files/port2_HW.cap"}
                  GET: /ixnetwork/files?filename=captures/packetCaptureFolder/port2_HW.cap
        """
        
        # For Linux API server
        if '\\' not in saveToTempLocation:
            # For Linux API server, need to give a name for a temporary folder
            saveToTempLocation = 'packetCaptureFolder'

            # For Linux API server, must get the vport name and cannot modify the vport name.
            vport = self.portMgmtObj.getVports([port])[0]
            response = self.ixnObj.get(self.ixnObj.httpHeader + vport)
            vportName = response.json()['name']

            vportName = vportName.replace('/', '_')
            vportName = vportName.replace(' ', '_')
            self.ixnObj.logInfo('vportName: {}'.format(vportName))

        if appendToSavedCapturedFile != None:
            data = {'arg1': saveToTempLocation, 'arg2': appendToSavedCapturedFile}
        else:
            data = {'arg1': saveToTempLocation}

        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/operations/savecapturefiles', data=data)
        self.ixnObj.waitForComplete(response, self.ixnObj.httpHeader+ response.json()['url'], timeout=300)

        if typeOfCapture == 'control':
            if '\\' not in saveToTempLocation:
                # For Linux
                capFileToGet = '/home/ixia_apps/web/platform/apps-resources/ixnetworkweb/configs/captures/{0}/{1}_SW.cap'.format(
                    saveToTempLocation, vportName)
            else:
                # For Windows
                capFileToGet = saveToTempLocation+'\\'+port[1]+'-'+port[2]+'_SW.cap'

        if typeOfCapture == 'data':
            if '\\' not in saveToTempLocation:
                # For Linux
                capFileToGet = '/home/ixia_apps/web/platform/apps-resources/ixnetworkweb/configs/captures/{0}/{1}_HW.cap'.format(
                    saveToTempLocation, vportName)
            else:
                capFileToGet = saveToTempLocation+'\\'+port[1]+'-'+port[2]+'_HW.cap'

        fileMgmtObj = FileMgmt(self.ixnObj)

        if self.ixnObj.serverOs in ['windows', 'windowsConnectionMgr']:
            fileMgmtObj.copyFileWindowsToLocalLinux(windowsPathAndFileName=capFileToGet, localPath=localLinuxLocation,
                                                    renameDestinationFile=None)
        if self.ixnObj.serverOs == 'linux':
            # Parse out captures/packetCaptureFolder/<vportName>_HW.cap
            match = re.search('.*(captures.*)', capFileToGet)
            pathExtension = match.group(1)

            fileMgmtObj.copyFileLinuxToLocalLinux(linuxApiServerPathAndFileName=capFileToGet, localPath=localLinuxLocation,
                                                  renameDestinationFile=None, linuxApiServerPathExtension=pathExtension)
            
